package jp.moke;

import java.io.IOException;
import java.net.UnknownHostException;

import pharerouge.hotline.BadMagicException;
import pharerouge.hotline.HLClient;
import pharerouge.hotline.HLClientAdapter;
import pharerouge.hotline.HLException;
import pharerouge.hotline.HLProtocol.UserListComponent;

public class HLFacade extends HLClientAdapter {

    HLClient hlClient;
    HLClientInterface hlcInterface;

    public void setHlcInterface(HLClientInterface hlcInterface) {
        this.hlcInterface = hlcInterface;
    }

    public void login(String host, int port, String id, String pass,
            String nick, HLClientInterface hlClientInterface)
            throws UnknownHostException, IOException, BadMagicException,
            HLException, InterruptedException {

        if (hlClient != null) {
            hlClient.close();
            hlClient = null;
        }

        hlClient = new HLClient(host, port, id, pass, nick, (short) 0);
        hlClient.setEncoding("Shift_JIS");
        this.setHlcInterface(hlClientInterface);
        hlClient.addHLClientListener(this);
        hlClient.connect();
        hlClient.login();
    }

    public void requestUserList() throws IOException {
        this.hlClient.requestUserList();
    }

    public void sendChat(String msg) throws IOException {
        if (this.hlClient != null) {
            hlClient.sendChat(msg);
        }
    }

    public void handleChat(String msg) {
        handleChat("Public", msg);
    }

    private void handleChat(Object domain, String msg) {

        delegateChat(domain, msg.replace('\r', '\n'));
    }

    private void delegateChat(Object domain, String msg) {
        this.hlcInterface.receiveMsg(msg);
    }

    public void close() throws IOException {
        if (hlClient != null) {
            hlClient.close();
            hlClient = null;
        }
    }

    @Override
    public void handleUserList(int id, UserListComponent[] uh) {

        this.hlcInterface.receiveUserList(uh);

    }

    public boolean isConnected() {
        return this.hlClient != null;
    }

    @Override
    public void handleUserLeave(int sock) {
        this.hlcInterface.manageUser(sock, null);
    }

    @Override
    public void handleUserChange(int sock, String nick, int icon, int color) {
        this.hlcInterface.manageUser(sock, nick);
    }
}
