package jp.moke;

import java.io.IOException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import pharerouge.hotline.BadMagicException;
import pharerouge.hotline.HLException;
import pharerouge.hotline.HLProtocol.UserListComponent;
import android.app.Activity;
import android.content.Intent;
import android.content.SharedPreferences;
import android.content.pm.ActivityInfo;
import android.os.Bundle;
import android.os.Handler;
import android.os.Message;
import android.preference.PreferenceManager;
import android.util.TypedValue;
import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewGroup;
import android.view.Window;
import android.widget.ArrayAdapter;
import android.widget.Button;
import android.widget.EditText;
import android.widget.ListView;
import android.widget.TextView;
import android.widget.Toast;

public class PharerougeMakaizouActivity extends Activity implements
        HLClientInterface {

    HLFacade hlfacade;

    Map<String, String> userlist;

    ListView listView;

    int fontsize;
    
    /** Called when the activity is first created. */
    @Override
    public void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        requestWindowFeature(Window.FEATURE_CUSTOM_TITLE);

        setContentView(R.layout.main);
        getWindow().setFeatureInt(Window.FEATURE_CUSTOM_TITLE, R.layout.title);
        // findViewById(id.titletext);

        SharedPreferences preferences = PreferenceManager
                .getDefaultSharedPreferences(this);
        if (preferences.getString(getString(R.string.pref_orientation),
                getString(R.string.pref_orientation_portrait)).equals(
                getString(R.string.pref_orientation_portrait))) {
            setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_PORTRAIT);

        } else {
            setRequestedOrientation(ActivityInfo.SCREEN_ORIENTATION_LANDSCAPE);
        }

        // editText1の幅を固定
        EditText editText = (EditText) findViewById(R.id.editText1);
        editText.setWidth(editText.getWidth());
        // scrollView = findViewById(R.id.scrollView1);
        // textView = (TextView) findViewById(R.id.textView1);
        Button button = (Button) findViewById(R.id.sendButton1);

        // ArrayAdapter<String> adapter = new ArrayAdapter<String>(this,
        // android.R.layout.simple_list_item_1);
        fontsize = Integer.parseInt(preferences.getString(
                getString(R.string.pref_fontsize), "18"));
        ArrayAdapter<String> adapter = new ArrayAdapter<String>(this,
                R.layout.listitem, R.id.listItemText) {
            public View getView(int position, View convertView, ViewGroup parent) {

                    
                TextView view = (TextView) super.getView(position, convertView,
                        parent);
                view.setTextSize(TypedValue.COMPLEX_UNIT_SP, (float) fontsize);
                return view;
            }

        };

        listView = (ListView) findViewById(R.id.listView1);

        listView.setAdapter(adapter);

        button.setOnClickListener(new View.OnClickListener() {
            public void onClick(View v) {

                EditText editText = (EditText) findViewById(R.id.editText1);
                if (editText.getText() != null
                        && editText.getText().length() > 0) {

                    try {
                        hlfacade.sendChat(editText.getText().toString());
                    } catch (NullPointerException e) {
                        // 文字列が未確定の場合発生する・・・がなぜか問題ない
                        e.printStackTrace();
                    } catch (IOException e) {
                        e.printStackTrace();
                        displayToast(getString(R.string.error_communication));
                        displayToast(e.toString());
                        close();
                    }
                    editText.setText("");
                }

            }
        });
    }

    // @Override
    // protected void onPause() {
    //
    // this.close();
    // }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        MenuInflater inflater = getMenuInflater();
        inflater.inflate(R.menu.menu, menu);

        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        switch (item.getItemId()) {
        case R.id.menuconnect:
            System.out.println("connect()");

            if (this.hlfacade != null) {
                close();
                displayToast(getString(R.string.message_disconnected));
            } else {
                connect();
            }
            break;
        case R.id.menupref:
            openPref();
            break;
        case R.id.bookmarkpref:
            openBookmarkPref();
            break;
        case R.id.userlistpref:
            displayUserlist();
            break;
        default:
            break;
        }

        return super.onOptionsItemSelected(item);
    }

	private void close() {

//		SharedPreferences preferences = PreferenceManager
//				.getDefaultSharedPreferences(this);
//		System.out.println(preferences.getBoolean(getString(R.string.pref_endurance), false));
//		if (!preferences.getBoolean(getString(R.string.pref_endurance), false)) {

			if (hlfacade != null) {

				try {
					hlfacade.close();
		            displayToast(getString(R.string.message_disconnected));
				} catch (IOException e) {
					e.printStackTrace();
					displayToast(e.toString());
				}
				this.hlfacade = null;
				findViewById(R.id.sendButton1).setEnabled(false);
				findViewById(R.id.editText1).setEnabled(false);
			}
			if (this.userlist != null) {
				userlist = null;
			}
			findViewById(R.id.imageView1).setVisibility(View.INVISIBLE);
//		}

	}

    private void connect() {

        SharedPreferences preferences = PreferenceManager
                .getDefaultSharedPreferences(this);
        String host = preferences.getString(getString(R.string.pref_url), "");
        try {
            int port = Integer.parseInt(preferences.getString(
                    getString(R.string.pref_port), "0"));
            String id = preferences.getString(getString(R.string.pref_id), "");
            String pass = preferences.getString(getString(R.string.pref_pass),
                    "");
            String nick = preferences.getString(getString(R.string.pref_nick),
                    "");
            if (host.length() == 0 || port == 0 || nick.length() == 0) {
                displayToast(getString(R.string.error_pref_invalid));
                return;
            }
            // textView.setTextSize(TypedValue.COMPLEX_UNIT_PT,
            // Float.parseFloat(fontsize));
            hlfacade = new HLFacade();

            hlfacade.login(host, port, id, pass, nick, this);
        } catch (ClassCastException e) {
            displayToast(getString(R.string.error_pref_invalid));
            e.printStackTrace();
            close();
            return;
        } catch (UnknownHostException e) {
            displayToast(getString(R.string.error_server_notfound));
            e.printStackTrace();
            close();
            return;
        } catch (IOException e) {
            displayToast(getString(R.string.error_connection_failed));
            e.printStackTrace();
            close();
            return;
        } catch (BadMagicException e) {
            displayToast(getString(R.string.error_invalid_response) + "\n" + e);
            e.printStackTrace();
            close();
            return;
        } catch (HLException e) {
            displayToast(getString(R.string.error_something) + "\n" + e);
            e.printStackTrace();
            close();
            return;
        } catch (InterruptedException e) {
            displayToast(getString(R.string.error_something));
            e.printStackTrace();
            close();
            return;
        }
        hlfacade.setHlcInterface(this);

        findViewById(R.id.sendButton1).setEnabled(true);
        findViewById(R.id.editText1).setEnabled(true);
        // ((TextView) findViewById(R.id.textView1)).setText("");
        ((ArrayAdapter<String>) listView.getAdapter()).clear();
        try {
            this.hlfacade.requestUserList();

        } catch (IOException e) {
            displayToast(getString(R.string.error_receive_userlist));
            e.printStackTrace();
        }
        findViewById(R.id.imageView1).setVisibility(View.VISIBLE);
        displayToast(getString(R.string.message_connected) + host);
    }

    @Override
    public boolean onPrepareOptionsMenu(Menu menu) {

        if (hlfacade == null) {
            menu.findItem(R.id.menuconnect).setTitle(
                    getString(R.string.menu_connect));
            menu.findItem(R.id.menuconnect).setIcon(
                    android.R.drawable.ic_media_play);
        } else {
            menu.findItem(R.id.menuconnect).setTitle(
                    getString(R.string.menu_disconnect));
            menu.findItem(R.id.menuconnect).setIcon(
                    android.R.drawable.sym_call_missed);
        }

        if (userlist == null) {
            menu.findItem(R.id.userlistpref).setEnabled(false);
        } else {
            menu.findItem(R.id.userlistpref).setEnabled(true);
        }

        return super.onPrepareOptionsMenu(menu);
    }

    private void openPref() {
        this.close();
        Intent intent = new Intent(this, ServerPrefActivity.class);
        startActivity(intent);
    }

    private void openBookmarkPref() {
        this.close();
        Intent intent = new Intent(this, BookmarkPrefActivity.class);
        startActivity(intent);
    }

    List<String> message = new ArrayList<String>();

    public synchronized void receiveMsg(String msg) {

        message.add(msg.trim());
    }

    long interval = 300;
    RepaintHandler handler = null;

    // View scrollView;

    class RepaintHandler extends Handler {
        public void handleMessage(Message msg) {
            synchronized (message) {
                if (message.size() != 0) {

                    ArrayAdapter<String> adapter = (ArrayAdapter<String>) listView
                            .getAdapter();
                    for (Iterator<String> iterator = message.iterator(); iterator
                            .hasNext();) {

                        adapter.add(iterator.next());

                    }

                    // textView.setText(textView.getText() + message);

                    message.clear();

                    // adapter.add("");
                    listView.setSelection(listView.getCount());
                    // listView.invalidate();
                    // adapter.remove(adapter.getItem(listView.getCount() - 1));
                    // scrollView.scrollTo(0, textView.getHeight());
                }
            }
            if (handler != null)
                handler.sleep(interval);
        }

        public void sleep(long delay) {
            removeMessages(0); // 既にあるMessageを削除
            sendMessageDelayed(obtainMessage(0), delay); // delayの時間後にMessageを発行する
        }
    }

    protected void onResume() {
        super.onResume();
        // 再描画を再開する
        handler = new RepaintHandler();
        handler.sleep(0);
    }

    protected void onPause() {
        super.onPause();
//        if (this.hlfacade != null) {
//
//            this.close();
//        }
        // 再描画を終了する
        handler = null;
    }

    private void displayToast(String msg) {
        Toast.makeText(this, msg, Toast.LENGTH_SHORT).show();

    }

    public void displayUserlist() {
        String msg = "";
        for (String key : this.userlist.keySet()) {
            msg += this.userlist.get(key) + "  ";
        }
        Toast.makeText(this, msg, Toast.LENGTH_LONG).show();
    }

    public void receiveUserList(UserListComponent[] users) {
        this.userlist = new HashMap<String, String>();
        for (UserListComponent userListComponent : users) {
            userlist.put(Integer.toString(userListComponent.sock),
                    userListComponent.nick);
        }
    }

    public void manageUser(int sock, String nick) {

        String sockStr = Integer.toString(sock);
        if (nick == null) {
            // delete
            this.receiveMsg("<" + userlist.get(sockStr)
                    + getString(R.string.message_logout) + ">");
            this.userlist.remove(sockStr);
        } else {
            // add
            if (!userlist.containsKey(sockStr)) {
                this.userlist.put(sockStr, nick);
                this.receiveMsg("<" + nick + getString(R.string.message_login)
                        + ">");
            }
        }
    }
    @Override
    protected void onDestroy() {
        this.close();
        super.onDestroy();
    }

    boolean exitflag = false;
    
    @Override
    public void onBackPressed() {
        if (exitflag) {
            this.close();
            super.onBackPressed();
            this.finish();
        }
        else {
            displayToast(getString(R.string.message_confirm_exit));
        }
        new Thread() {
            public void run() {
                try {
                    exitflag = true;
                    sleep(5000);
                    exitflag = false;
                    
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
            }
        }.start();

    }
}