/**
 *  $Id: HLClientDispatcher.java,v 1.16 2001/10/08 22:03:28 groomed Exp $
 *  Copyright (C) 1998-2001 groomed <groomed@users.sourceforge.net> This program
 *  is free software; you can redistribute it and/or modify it under the terms
 *  of the GNU General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or (at your option) any later
 *  version. This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 *  more details. You should have received a copy of the GNU General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package pharerouge.hotline;

import java.io.IOException;
import java.util.Calendar;
import java.util.Enumeration;
import java.util.Vector;

import pharerouge.utils.DebuggerOutput;
import pharerouge.utils.ToArrayConverters;

/**
 *  This class is used by the HLClient class for the handling of incoming
 *  events.
 */
public class HLClientDispatcher {

    HLProtocol hlp;
    HLClient hlc;
    boolean agreementAccepted = true, agreementAborted = false, 
        agreementReceived = false;

    /* Experimentally added 2002.03.01 MackyBee */
    boolean waitingAgree;
    /* Experimentally added 2002.03.04 MackyBee */
    boolean waitNoTimeLimit;

    /**
     *  Constructor for the HLClientDispatcher object
     *
     * @param  h  Description of the Parameter
     */
    HLClientDispatcher(HLClient h) {
        hlc = h;
        // 2002.12.22 MackyBee EXPERIMENTALLY modified. 
        hlp = hlc.getHLProtocol();
        //hlp = new HLProtocol();
    }
    
    // 2002.12.22 MackyBee added. 
    HLClientDispatcher(HLClient h, String encoding) {
        hlc = h;
        hlp = new HLProtocol(encoding);
    }
    
    // 2002.12.22 MackyBee added. 
    //void setEncoding(String encoding) {
    //    
    //}

    /**
     *  Main dispatcher.
     *
     * @param  packet  the packet to process.
     */
    public void dispatch(HLProtocol.Packet packet) {

        try {
            switch (packet.header.id) {
                case HLProtocol.HTLS_HDR_TASK:
                    dispatchTask(packet);
                    break;
                case HLProtocol.HTLS_HDR_CHAT:
                    dispatchChat(packet);
                    break;
                case HLProtocol.HTLS_HDR_MSG:
                    dispatchMessage(packet);
                    break;
                case HLProtocol.HTLS_HDR_USER_CHANGE:
                    dispatchUserChange(packet);
                    break;
                case HLProtocol.HTLS_HDR_ACCOUNT_INFO:
                    dispatchAccountInfo(packet);
                    break;
                case HLProtocol.HTLS_HDR_USER_LEAVE:
                    dispatchUserLeave(packet);
                    break;
                case HLProtocol.HTLS_HDR_NEWS_POST:
                    dispatchNewsPost(packet);
                    break;
                case HLProtocol.HTLS_HDR_CHAT_SUBJECT:
                    dispatchPrivateChatSubject(packet);
                    break;
                case HLProtocol.HTLS_HDR_CHAT_INVITE:
                    dispatchPrivateChatInvite(packet);
                    break;
                case HLProtocol.HTLS_HDR_CHAT_USER_CHANGE:
                    dispatchPrivateChatUserChange(packet);
                    break;
                case HLProtocol.HTLS_HDR_CHAT_USER_LEAVE:
                    dispatchPrivateChatUserLeave(packet);
                    break;
                case HLProtocol.HTLS_HDR_AGREEMENT:
                    dispatchAgreement(packet);
                    break;
                case HLProtocol.HTLS_HDR_QUEUEPOS:
                    dispatchTransferQueuePos(packet);
                    break;
                default:
                    DebuggerOutput.debug("dispatch: unknown header type: 0x" 
                        + Integer.toHexString(packet.header.id));
            }
        } catch (IOException e) {
            // 2002.08.26 MackyBee remained.
            DebuggerOutput.stackTrace(e);
        } catch (HLTaskNotFoundException e) {
            DebuggerOutput.stackTrace(e);
        } catch (Exception e) {
            // 2002.12.21 MackyBee EXPERIMENTALLY added. 
            e.printStackTrace();
        }
    }

    /* Private chat */
    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchPrivateChatUserChange(HLProtocol.Packet packet) 
        throws IOException {

        int icon = 0;

        int sock = 0;

        int color = 0;

        int pcref = 0;
        byte[] nick = new byte[0];

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {

                case HLProtocol.HTLS_DATA_SOCKET:
                    sock = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_ICON:
                    icon = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_NICK:
                    nick = dh.data;
                    break;
                case HLProtocol.HTLS_DATA_COLOUR:
                    color = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_CHAT_REF:
                    pcref = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
            }

        }

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handlePrivateChatUserChange(pcref, sock, 
                    // 2002.12.28 MackyBee modified. 
                    new String(nick, hlc.encoding), 
                    //new String(nick), 
                    icon, color);
        }

    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchPrivateChatUserLeave(HLProtocol.Packet packet) 
        throws IOException {

        int sock = 0;

        int pcref = 0;

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {

                case HLProtocol.HTLS_DATA_CHAT_REF:
                    pcref = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_SOCKET:
                    sock = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
            }

        }

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handlePrivateChatUserLeave(pcref, sock);
        }
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchPrivateChatSubject(HLProtocol.Packet packet) 
        throws IOException {

        int pcref = 0;
        byte[] subject = new byte[0];

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {

                case HLProtocol.HTLS_DATA_CHAT_REF:
                    pcref = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_CHAT_SUBJECT:
                    subject = dh.data;
                    break;
            }

        }

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handlePrivateChatSubject(pcref, 
                    // 2002.12.28 MackyBee modified. 
                    new String(subject, hlc.encoding));
                    //new String(subject));
        }
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchPrivateChatInvite(HLProtocol.Packet packet) 
        throws IOException {

        int sock = 0;

        int pcref = 0;
        byte[] nick = new byte[0];

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {

                case HLProtocol.HTLS_DATA_SOCKET:
                    sock = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_CHAT_REF:
                    pcref = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_NICK:
                    nick = dh.data;
                    break;
            }

        }

        if (nick != null) {
            for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
                ((HLClientListener) en.nextElement()).
                    handlePrivateChatInvite(pcref, sock, 
                    // 2002.12.28 MackyBee modified. 
                    new String(nick, hlc.encoding));
                    //new String(nick));
            }
        }
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    Object dispatchPrivateChatCreate(HLProtocol.Packet packet) throws IOException {

        int sock = 0;

        int pcref = 0;

        int icon = 0;

        int color = 0;
        byte[] nick = new byte[0];

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {

                case HLProtocol.HTLS_DATA_SOCKET:
                    sock = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_ICON:
                    icon = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_NICK:
                    nick = dh.data;
                    break;
                case HLProtocol.HTLS_DATA_COLOUR:
                    color = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_CHAT_REF:
                    pcref = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                default:
                    DebuggerOutput.debug("dispatchPrivateChatCreate: unknown dh type: " + dh.type);
                    break;
            }

        }

        DebuggerOutput.debug("calling listener for private chat create");
        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handlePrivateChatCreate(pcref, sock, 
                    // 2002.12.28 MackyBee modified. 
                    new String(nick, hlc.encoding), 
                    //new String(nick), 
                    icon, color);
        }

        return null;
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    Object dispatchPrivateChatJoin(HLProtocol.Packet packet) throws IOException {

        Vector userV = new Vector();

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {

                case HLProtocol.HTLS_DATA_USER_LIST:
                    userV.addElement((HLProtocol.UserListComponent) dh);
                    break;
                default:
                    DebuggerOutput.debug("dispatchPrivateChatJoin: unknown dh type: " + dh.type);
                    break;
            }

        }

        HLProtocol.UserListComponent[] users =
            new HLProtocol.UserListComponent[userV.size()];
        for (int i = 0; i < userV.size(); i++) {
            users[i] = (HLProtocol.UserListComponent) userV.elementAt(i);
        }
        userV.removeAllElements();

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handlePrivateChatJoin(packet.header.trans, users);
        }

        return users;
    }


    /* Chat, messagse */
    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchChat(HLProtocol.Packet packet) throws IOException {

        byte[] msg = new byte[0];
        
        /* 2002.04.20 MackyBee modified.
           -1 may be used by PR server. */
        int pcref = 0;
        //int pcref = -1;

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();
            
            switch (dh.type) {
                case HLProtocol.HTLC_DATA_CHAT_REF:
                    pcref = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLC_DATA_CHAT:
                    msg = dh.data;
                    break;
            }

        }

        if (msg != null) {
            /* 2002.05.02 MackyBee added temporary */
            //DebuggerOutput.putHexString(msg);
            
            /* 2002.04.20 MackyBee modified.
               Strictly, pcref <= 0 is unusual value. So, modified. 
               -> only 0. under 0 values are used by PR server. */
            if (pcref == 0) {
            //if (pcref <= 0) {
            //if (pcref == -1) {
                for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
			// 2006-03-08 Gardien added support for the <ver> command.
			HLClientListener cl = (HLClientListener) en.nextElement();
			String someString = new String(msg, hlc.encoding);
                    	cl.handleChat(someString);
			//if (someString.lastIndexOf("<ver>") == someString.length() - 1 - 5){
			//if (someString.length() == 22 && someString.indexOf(":  <ver>") != -1){	// 22 = " (1)nick(13):  (3)<ver>(5)"
			/* This line should be changed to trigger the <ver> command only on lines that exatcly match it 
				(instead of anywhere in the middle of a sentence), but the problem is that the way a chat message is 
				echoed differs from a server to another... */
			if (someString.indexOf("<ver>") != -1){
				cl.handleVer();
			}
                }
            } else {
                for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
                    ((HLClientListener) en.nextElement()).
                        handlePrivateChat(pcref, 
                        // 2002.12.28 MackyBee modified. 
                        new String(msg, hlc.encoding));
                        //new String(msg));
                }
            }
        }
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchMessage(HLProtocol.Packet packet) throws IOException {

        byte[] msg = new byte[0];
        byte[] nick = null;
        int sock = 0;

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {
                case HLProtocol.HTLC_DATA_SOCKET:
                    sock = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLC_DATA_MSG:
                    msg = dh.data;
                    break;
                case HLProtocol.HTLC_DATA_NICK:
                    nick = dh.data;
                    break;
                default:
                    DebuggerOutput.debug(
                        "dispatchMessage: unknown dh.type: 0x" 
                        + Integer.toHexString(dh.type));
                    break;
            }

        }

        if (nick == null) {
            for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
                ((HLClientListener) en.nextElement()).
                    // 2002.12.28 MackyBee modified. 
                    handleAdministratorMessage(new String(msg, hlc.encoding));
                    //handleAdministratorMessage(new String(msg));
            }
        } else {
            for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
                ((HLClientListener) en.nextElement()).
                    handleMessage(sock, 
                        // 2002.12.28 MackyBee modified. 
                        new String(nick, hlc.encoding), 
                        //new String(nick), 
                        // 2002.12.28 MackyBee modified. 
                        new String(msg, hlc.encoding));
                        //new String(msg));
            }
        }
    }

    /* Files */
    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchTransferQueuePos(HLProtocol.Packet packet) 
        throws IOException {

        int ref = -1;
        int queuepos = -1;

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {

                case HLProtocol.HTLS_DATA_HTXF_REF:
                    ref = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_HTXF_QUEUEPOS:
                    queuepos = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
            }

        }

        if (ref != -1 && queuepos != -1) {
            for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
                ((HLClientListener) en.nextElement()).
                    handleTransferQueuePosition(ref, queuepos);
            }
        }

    }


    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    Object dispatchFileGet(HLProtocol.Packet packet) throws IOException {

        /* Note 2002.03.07 MackyBee */
        /* 	The resume switch may be as well like static variable,
			because users have to wait every end of file downloading. */

//        long size = 0;
//        long fileSize = 0;
//        int ref = 0;
//        int queuepos = -1;
//
//        /* 2002.05.31 MackyBee modified */
//        HLProtocol.DataComponent dh;
//        
//        while (packet.hasMoreComponents()) {
//
//            /* 2002.05.31 MackyBee modified */
//            dh = packet.nextComponent();
//            //HLProtocol.DataComponent dh = packet.nextComponent();
//
//            switch (dh.type) {
//                case HLProtocol.HTLS_DATA_HTXF_SIZE:
//                    size = ToArrayConverters.byteArrayToInt(dh.data);
//                    break;
//                case HLProtocol.HTLS_DATA_FILE_SIZE:
//                    fileSize = ToArrayConverters.byteArrayToInt(dh.data);
//                    break;
//                case HLProtocol.HTLS_DATA_HTXF_REF:
//                    ref = ToArrayConverters.byteArrayToInt(dh.data);
//                    break;
//                case HLProtocol.HTLS_DATA_HTXF_QUEUEPOS:
//                    queuepos = ToArrayConverters.byteArrayToInt(dh.data);
//                    break;
//            }
//
//        }
//
//        HLProtocol hlp = new HLProtocol();
//
//        Socket xfsocket = null;
//        Transferrer transferrer = null;
//        PipedOutputStream outputPipe = null;
//        Transfer xf = hlc.getTransfer(packet.header.trans);
//
//        if (xf == null) {
//            throw new IOException("Invalid transfer ID " 
//                + packet.header.trans);
//        }
//
//        try {
//
//            DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans + 
//                "]: creating Transferrer");
//            transferrer = new Transferrer(packet.header.trans, ref, xf.meter);
//
//            if (xf.forkOrFile == HLProtocol.DL_DATA_FORK) {
//                /* Get just the data fork, no headers, and stream that
//                   data to a pipe. */
//                DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                    + "]: creating output pipe");
//                outputPipe = new PipedOutputStream();
//                xf.inputPipe.connect(outputPipe);
//            }
//
//            DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                + "]: starting Meter");
//
//            xf.meter.startMeter(transferrer, ref, xf.name, size);
//
//            if (queuepos != -1) {
//                /* If we got a queue parameter, then tell the Meter
//                   we're about to start, and immediately notify the
//                   event handler about the queue position. */
//                DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                    + "]: got queuepos = " + queuepos 
//                    + ", calling handleTransferQueuePosition");
//
//                for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
//                    ((HLClientListener) en.nextElement()).
//                        handleTransferQueuePosition(ref, queuepos);
//                }
//            }
//
//            DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans + 
//                "]: opening transfer socket");
//
//            xfsocket = new Socket(hlc.getHost(), hlc.getPort() + 1);
//
//            /* Disabled 2002.03.07 MackyBee */
//            //xfsocket.setSoTimeout(500);
//            //xfsocket.setSoLinger(true, 5);
//
//            // 2003.01.08 MackyBee deactivated. 
//            //xfsocket.setReceiveBufferSize(HLProtocol.SOCK_RECEIVE_BUF_SIZE);
//            
//            // 2003.01.08 MackyBee added. 
//            transferrer.setBufferSize(xfsocket.getReceiveBufferSize());
//
//            DataOutputStream xfoutput =
//                new DataOutputStream(new BufferedOutputStream(
//                xfsocket.getOutputStream()));
//            /* 2002.05.11 MackyBee modified.
//               BufferedInputStream has a bug (ID: 4479751). 
//               And experimentally removes InterruptableInputStream wrapping. 
//               Result: Became smooth, and no problem. */
//            DataInputStream xfinput = 
//                new DataInputStream(xfsocket.getInputStream());
//                //new DataInputStream(new InterruptableInputStream(
//                //xfsocket.getInputStream()));
//                //new DataInputStream(new BufferedInputStream(
//                //new InterruptableInputStream(xfsocket.getInputStream())));
//
//            DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                + "]: setting up Transferrer");
//
//            transferrer.setDataInputStream(xfinput);
//            transferrer.setDataOutputStream(xfoutput);
//
//            /* Send the file transfer header and receive the file
//               transfer info in reply. */
//            HLProtocol.FileTransferHeader fth =
//                hlp.new FileTransferHeader(ref, 0);
//
//            fth.write(xfoutput);
//
//            HLProtocol.FileTransferInfo fti = null;
//
//            if (xf.forkOrFile != HLProtocol.DL_DATA_FORK) {
//
//                // Note: Actually, reads stream. Then, can't omit.
//                fti = hlp.new FileTransferInfo(xfinput);
//
//                // 2002.09.06 MackyBee added.
//                if (xf.local.getDataSize() == 0) {
//                    if (xf.local instanceof SplitMacFile) {
//                        try {
//                            if (pharerouge.client.Main.rlo.getBooleanProperty(
//                                "Toggle.CreateMetaInfoFile")) {
//                                writeFileTransferInfo(xf.local, fti);
//                            }
//                        } catch (Exception e) {}
//                    } else {
//                        writeFileTransferInfo(xf.local, fti);
//                    }
//                }
//
//                // 2002.09.06 MackyBee modified this process to above.
//                /* Initialize the local file. */
//                /*
//                if (xf.local.getDataSize() == 0) {
//                    xf.local.writeHeader(fti.fileName,
//                        fti.fileType,
//                        fti.fileCreator,
//                        fti.fileComment,
//                        fti.creationDate,
//                        fti.modificationDate,
//                        fti.finderFlags);
//                }
//                */
//            }
//
//            byte[] l = null;
//
//            if (xf.forkOrFile != HLProtocol.DL_DATA_FORK) {
//
//                /* Get the data fork size. Maybe should check whether
//                   magic matches something (e.g. "DATA"). This reads the
//                   file size as a 64 bit long integer. */
//                l = new byte[8];
//                xfinput.skipBytes(8);
//                xfinput.readFully(l, 0, 8);
//                size = ToArrayConverters.byteArrayToLong(l);
//
//                /* Resume. */
//                xf.local.getDataFork().seek(xf.local.getDataFork().size());
//
//            }
//
//            /* Transfer the data fork. */
//            DataOutput output;
//
//            /* Setup the output to go to a file for downloads or
//               to a pipe for viewing. */
//            if (xf.forkOrFile != HLProtocol.DL_DATA_FORK) {
//                output = xf.local.getDataFork().getDataOutput();
//            } else {
//                output = new DataOutputStream(outputPipe);
//            }
//
//            try {
//                DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                    + "]: transferring data fork ...");
//                transferrer.transferDataFork(xfinput, output, size);
//            } catch (InterruptedException e) {
//                throw e;
//            } catch (IOException e) {
//                throw e;
//            } finally {
//                if (xf.forkOrFile != HLProtocol.DL_DATA_FORK) {
//                    DebuggerOutput.debug("dispatchFileGet[" 
//                        + packet.header.trans + "]: closing data fork.");
//                    xf.local.setDataSize(xf.local.getDataSize() 
//                        + transferrer.getDataBytesWritten());
//                    xf.local.getDataFork().close();
//                }
//            }
//
//            if (xf.forkOrFile != HLProtocol.DL_DATA_FORK) {
//
//                size = 0;
//
//                if (fti.numberOfBlocks == 3) {
//                    /* Should check for MACR. */
//                    xfinput.skipBytes(8);
//                    xfinput.readFully(l, 0, 8);
//                    size = ToArrayConverters.byteArrayToLong(l);
//                }
//
//                /* Resume. */
//                xf.local.getResourceFork().seek(xf.local.getResourceFork().
//                    size());
//
//                /* Transfer the resource fork. */
//                try {
//
//                    DebuggerOutput.debug("dispatchFileGet[" 
//                        + packet.header.trans 
//                        + "]: transferring resource fork");
//
//                    transferrer.transferResourceFork(xfinput,
//                        xf.local.getResourceFork().
//                        getDataOutput(), size);
//
//                } catch (InterruptedException e) {
//                    throw e;
//                } catch (IOException e) {
//                    throw e;
//                } finally {
//                    DebuggerOutput.debug("dispatchFileGet[" 
//                        + packet.header.trans + "]: closing resource fork");
//                    xf.local.setResourceSize(xf.local.getResourceSize() 
//                        + transferrer.getResourceBytesWritten());
//                    xf.local.getResourceFork().close();
//                }
//
//                /* Tell the Meter that the transmission is over. */
//            }
//
//            DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                + "]: stopping meter");
//            xf.meter.stopMeter(ref);
//
//            if (xf.forkOrFile != HLProtocol.DL_DATA_FORK) {
//                /* Set MacOS attributes on file if possible. */
//                MacFileUtils.setFileTypeAndCreator(
//                    xf.local.getFile(), 
//                    // 2002.12.04 MackyBee modified. 
//                    fti.fileType.getBytes(hlc.encoding), 
//                    //fti.fileType.getBytes(), 
//                    // 2002.12.04 MackyBee modified. 
//                    fti.fileCreator.getBytes(hlc.encoding));
//                    //fti.fileCreator.getBytes());
//            }
//        } catch (InterruptedException e) {
//            /* This doesn't need handling, it is thrown when the
//               Transferrer is interrupted by it's Meter. */
//            DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                + "]: caught InterruptedException");
//        } catch (SocketException e) {
//            // 2002.07.04 MackyBee EXPERIMENTALLY added. 
//            // I think if downloading data reached to its end, this exception 
//            // will be occurred naturally. -> 2002.09.01 Note: Not natural.
//            // Result: Usually, this exception is not caughten.
//            DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                + "]: caught SocketException: " + e.getMessage());
//
//            // 2002.09.01 MackyBee experimentally added.
//            if (!transferrer.interrupted) {
//                xf.meter.stopMeterWithError(ref, e);
//            }
//
//        } catch (IOException e) {
//            DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                + "]: caught IOException: " + e.toString());
//            /* 2002.07.01 MackyBee disabled. */
//            //DebuggerOutput.stackTrace(e);
//
//            if (!transferrer.interrupted) {
//                xf.meter.stopMeterWithError(ref, e);
//            }
//        } finally {
//
//            DebuggerOutput.debug("dispatchFileGet[" + packet.header.trans 
//                + "]: in final block");
//
//            try {
//
//                if (outputPipe != null) {
//                    outputPipe.flush();
//                    outputPipe.close();
//                }
//
//                if (xf.forkOrFile != HLProtocol.DL_DATA_FORK) {
//                    if (xf.local != null) {
//                        xf.local.cleanup();
//                    }
//                }
//
//                if (xfsocket != null) {
//                    DebuggerOutput.debug("dispatchFileGet[" 
//                        + packet.header.trans + "]: closing socket");
//                    xfsocket.close();
//                }
//            } catch (IOException e) {}
//
//        }

        return null;
    }

    /**
     * @author MackyBee
     * @since 2002.09.06
     */
//    private void writeFileTransferInfo(MacFile local, 
//        HLProtocol.FileTransferInfo fti) throws IOException {
//
//        local.writeHeader(
//            fti.fileName,
//            fti.fileType,
//            fti.fileCreator,
//            fti.fileComment,
//            fti.creationDate,
//            fti.modificationDate,
//            fti.finderFlags);
//    }

//    /**
//     * Dispatches received file uploading request.
//     *
//     * @param  packet           Description of the Parameter
//     * @return                  Description of the Return Value
//     * @exception  IOException  Description of the Exception
//     * @author Pascal Haarkmat [Original]
//     */
//    Object dispatchFilePut(HLProtocol.Packet packet) throws IOException {
//        // 2002.07.19 MackyBee added the method description.
//
//        // 2002.07.22 MackyBee note: 
//        // IMPORTANT: packet.header.isError handling is acted in dispatchTask
//        // method. 
//
//        int size = 0;
//        int ref = 0;
//
//        HLProtocol.ResumeTransferComponent rflt = 
//            hlp.new ResumeTransferComponent();
//
//        /* 2002.05.31 MackyBee modified */
//         HLProtocol.DataComponent dh;
//        
//        while (packet.hasMoreComponents()) {
//
//            /* 2002.05.31 MackyBee modified */
//            dh = packet.nextComponent();
//            //HLProtocol.DataComponent dh = packet.nextComponent();
//
//            switch (dh.type) {
//                case HLProtocol.HTLS_DATA_HTXF_REF:
//                    ref = ToArrayConverters.byteArrayToInt(dh.data);
//                    break;
//                case HLProtocol.HTLS_DATA_HTXF_RFLT:
//                    rflt = (HLProtocol.ResumeTransferComponent) dh;
//                    break;
//                default:
//                    DebuggerOutput.debug("dispatchFilePut: unknown dh.type: " 
//                        + dh.type);
//                    break;
//            }
//
//        }
//
//        Transferrer transferrer = null;
//        DataOutputStream xfoutput = null;
//        Transfer xf = hlc.getTransfer(packet.header.trans);
//
//        if (xf == null) {
//            throw new IOException("Invalid transaction ID " 
//                + packet.header.trans);
//        }
//
//        Socket xfsocket = null;
//
//        long rflt_data_size = ((Long) rflt.chunks.get("DATA")).longValue();
//        long rflt_macr_size = ((Long) rflt.chunks.get("MACR")).longValue();
//        long dataRemaining = xf.local_data_size - rflt_data_size;
//        long rsrcRemaining = xf.local_rsrc_size - rflt_macr_size;
//
//        // 2002.09.07 MackyBee TEMPORARY added.
//        DebuggerOutput.debug("dispatchFilePut [rflt_data_size]: " 
//            + rflt_data_size);
//        DebuggerOutput.debug("dispatchFilePut [rflt_macr_size]: " 
//            + rflt_macr_size);
//        DebuggerOutput.debug("dispatchFilePut [dataRemaining]: " 
//            + dataRemaining);
//        DebuggerOutput.debug("dispatchFilePut [rsrcRemaining]: " 
//            + rsrcRemaining);
//
//        // 2002.09.07 MackyBee moved this from the following try block.
//        if (dataRemaining < 0 || rsrcRemaining < 0) {
//            throw new IOException("Attempt to resume beyond end of file");
//        }
//
//        try {
//            /* Tell the Transferrer we're starting the
//               transmission. */
//            transferrer = new Transferrer(packet.header.trans, ref, xf.meter);
//            xf.meter.startMeter(transferrer,
//                ref,
//                xf.local.getFile().getName(),
//                dataRemaining + rsrcRemaining);
//                
//            //if (dataRemaining < 0 || rsrcRemaining < 0) {
//            //    throw new IOException(
//            //        "Attempt to resume beyond end of file");
//            //}
//
//            xfsocket = new Socket(hlc.getHost(), hlc.getPort() + 1);
//
//            // 2002.03.15 MackyBee
//            // Disabled for obscure intentions
//            //xfsocket.setSoTimeout(2500);
//            //xfsocket.setSoLinger(true, 5);
//
//            // 2003.01.08 MackyBee deactivated. 
//            //xfsocket.setSendBufferSize(HLProtocol.SOCK_SEND_BUF_SIZE);
//            
//            // 2003.01.08 MackyBee added. 
//            transferrer.setBufferSize(xfsocket.getSendBufferSize());
//
//            xfoutput = new DataOutputStream(xfsocket.getOutputStream());
//            // 2002.07.19 MackyBee EXPERIMENTALLY modified.
//            // Result: At least, uploading speed of offline is increased.
//            // Maybe, online also.
//            DataInputStream xfinput = 
//                new DataInputStream(xfsocket.getInputStream());
//                //new DataInputStream(new InterruptableInputStream(
//                //xfsocket.getInputStream()));
//
//            transferrer.setDataInputStream(xfinput);
//            transferrer.setDataOutputStream(xfoutput);
//
//            // 2002.07.26 MackyBee note: fti is used in the next 'fth'.
//            /* Construct the FILP file transfer info.
//               Changes here mean changes in HLClient.java. */
//            HLProtocol.FileTransferInfo fti =
//                hlp.new FileTransferInfo(xf.local.getFile().getName(),
//                xf.local.getType(),
//                xf.local.getCreator(),
//                xf.local.getComment(),
//                xf.local.getCreationDate(),
//                xf.local.getModificationDate(),
//                xf.local.getFinderFlags());
//
//            /* Construct the HTXF file upload header
//               (16 = size of HTXF header). */
//            HLProtocol.FileTransferHeader fth =
//                hlp.new FileTransferHeader(ref, 
//                // 2002.11.24 MackyBee re-modified. NO, here should be 16.
//                // 2002.07.26 MackyBee replaced.
//                // According to official client behavour, here should be 32.
//                //32 + 
//                16 +
//                (long) (dataRemaining +
//                rsrcRemaining +
//                fti.data.length));
//
//            /* Write the HTXF and FILP headers. */
//            fth.write(xfoutput);
//            fti.write(xfoutput);
//            xfoutput.flush();
//
//            // 2002.09.07 MackyBee added TEMPORARY.
//            //DebuggerOutput.debug("dispatchFilePut: Process 1 passed.");
//
//            // 2002.09.07
//            // MackyBee added if statement.
//            // Write the DATA block. Use a long for the file size.
//            if (dataRemaining > 0) {
//                // 2002.12.04 MackyBee modified. 
//                xfoutput.write(new String("DATA").getBytes(hlc.encoding));
//                //xfoutput.write(new String("DATA").getBytes());
//                xfoutput.writeInt(0);
//                xfoutput.writeLong(dataRemaining);
//                xfoutput.flush();
//            }
//
//            // 2002.09.07 MackyBee added TEMPORARY.
//            //DebuggerOutput.debug("dispatchFilePut: Process 2 passed.");
//
//            /* Transmit the data fork. */
//            if (dataRemaining > 0) {
//                xf.local.getDataFork().seek(rflt_data_size);
//                transferrer.transferDataFork(xf.local.getDataFork().
//                getDataInput(), xfoutput, dataRemaining);
//            }
//
//            // 2002.09.07 MackyBee added TEMPORARY.
//            //DebuggerOutput.debug("dispatchFilePut: Process 3 passed.");
//
//            // 2002.09.07
//            // MackyBee added if statement.
//            // Write the MACR block.
//            if (rsrcRemaining > 0) {
//                // 2002.12.04 MackyBee modified. 
//                xfoutput.write(new String("MACR").getBytes(hlc.encoding));
//                //xfoutput.write(new String("MACR").getBytes());
//                xfoutput.writeInt(0);
//                xfoutput.writeLong(rsrcRemaining);
//                xfoutput.flush();
//            }
//
//            // 2002.09.07 MackyBee added TEMPORARY.
//            //DebuggerOutput.debug("dispatchFilePut: Process 4 passed.");
//
//            /* Transmit the resource fork. */
//            if (rsrcRemaining > 0) {
//                xf.local.getResourceFork().seek(rflt_macr_size);
//                transferrer.transferResourceFork(xf.local.getResourceFork().
//                    getDataInput(), xfoutput, rsrcRemaining);
//            }
//
//            // 2002.09.07 MackyBee added TEMPORARY.
//            //DebuggerOutput.debug("dispatchFilePut: Process 5 passed.");
//
//            // 2002.09.07 MackyBee deactivated for unneeded.
//            //xfoutput.flush();
//
//            // 2002.09.07
//            // MackyBee experimentally deactivated this process.
//            // Because unusual process.
//            /*
//            if (!HLProtocol.asynchronousFileTransfers) {
//                try {
//                    // Wait for the server to close the connection.
//                    xfinput.read();
//                } catch (IOException e) {}
//            }
//            */
//
//            /* Tell the Meter that the transmission
//               has ended. */
//            xf.meter.stopMeter(ref);
//
//        } catch (InterruptedException e) {
//            /* This doesn't need handling, it is thrown when the
//               Transferrer is interrupted by it's Meter. */
//        } catch (SocketException e) {
//            // 2002.07.19 MackyBee EXPERIMENTALLY added.
//            DebuggerOutput.debug("dispatchFilePut[" + packet.header.trans 
//                + "]: caught SocketException: " + e.getMessage());
//
//            // 2002.09.07
//            // MackyBee deactivated data size checking process because this 
//            // way can not measure the size correctly.
//            // 2002.09.07
//            // MackyBee added data size checking process.
//            // 2002.08.30
//            // MackyBee modified experimentally.
//            // I think this should call 'stopMeterWithError'.
//            // 2002.07.20 MackyBee note: 
//            // I'd like to add a file size check process before stopping 
//            // the downloading process.
//            // 2002.07.20 MackyBee EXPERIMENTALLY added this.
//            // Result: Properly worked.
//            //if ((dataRemaining > 0) || (rsrcRemaining > 0)) {
//                xf.meter.stopMeterWithError(ref, e);
//            //} else {
//            //    DebuggerOutput.debug("dispatchFilePut[" + packet.header.trans 
//            //    + "]: caught SocketException, but usual.");
//            //}
//            //xf.meter.stopMeter(ref);
//
//        } catch (IOException e) {
//            // 2002.09.07 MackyBee EXPERIMENTALLY added.
//            DebuggerOutput.debug("dispatchFilePut[" + packet.header.trans 
//                + "]: caught IOException: " + e.getMessage());
//
//            // 2002.09.07
//            // MackyBee deactivated data size checking process because this 
//            // way can not measure the size correctly.
//            // 2002.09.07
//            // MackyBee added data size checking process.
//            //if ((dataRemaining > 0) || (rsrcRemaining > 0)) {
//                xf.meter.stopMeterWithError(ref, e);
//            //} else {
//            //     DebuggerOutput.debug("dispatchFilePut[" + packet.header.
//            // trans 
//            //    + "]: caught IOException, but usual.");
//            //}
//
//        } finally {
//            try {
//                if (xfoutput != null) {
//                    xfoutput.close();
//                }
//
//                if (xfsocket != null) {
//                    xfsocket.close();
//                }
//            } catch (IOException e) {}
//        }
//
//        return null;
//    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    Object dispatchFileList(HLProtocol.Packet packet) throws IOException {

        Vector fileV = new Vector();

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {
            dh = packet.nextComponent();

            switch (dh.type) {
                case HLProtocol.HTLS_DATA_FILE_LIST:
                    fileV.insertElementAt((HLProtocol.FileListComponent) dh, 0);
                    break;
            }
        }

        HLProtocol.FileListComponent[] files =
            new HLProtocol.FileListComponent[fileV.size()];

        for (int i = 0; i < fileV.size(); i++) {
            files[i] = (HLProtocol.FileListComponent) fileV.elementAt(i);
        }
        fileV.removeAllElements();

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handleFileList(packet.header.trans, files);
        }

        return files;
    }

    /* 2002.05.27 MackyBee experimentally added.
       Note: It may doesn't need to use Vector. "Packet.header.hc" keeps an 
       exact number of components value. */
    Object dispatchNewsBundleList(HLProtocol.Packet packet) 
        throws IOException {

        Vector listVector = new Vector();

        HLProtocol.NewsBundleListComponent tempComp;
        
        while (packet.hasMoreComponents()) {
            tempComp = (HLProtocol.NewsBundleListComponent)packet.
                nextComponent();
            
            switch (tempComp.type) {
                case HLProtocol.HTLS_DATA_BUNDLE_LIST:
                    listVector.addElement(tempComp);
                    break;
            }
        }

        HLProtocol.NewsBundleListComponent[] bundles =
            new HLProtocol.NewsBundleListComponent[listVector.size()];

        for (int i = 0; i < listVector.size(); i++) {
            bundles[i] = (HLProtocol.NewsBundleListComponent)listVector.
                elementAt(i);
        }
        listVector = null;
        
        /* Provisionally disabled */
        /*
        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handleNewsBundleList(packet.header.trans, bundles);
        }
        */
        return bundles;
    }

    /* 2002.06.04 MackyBee added. 
       Experimental spec. */
    protected Object dispatchNewsThreadList(HLProtocol.Packet packet) 
        throws IOException {

        Object tempObj = null;
        //while (packet.hasMoreComponents()) {
            tempObj = packet.nextComponent();
        //}
        return (tempObj);
    }
    
    /* 2002.06.05 MackyBee */
    protected Object dispatchNewsArticle(HLProtocol.Packet packet) 
        throws IOException {

        String article = null;
        //HLProtocol.NewsArticleInfo articleInfo = hlp.new NewsArticleInfo();

        HLProtocol.DataComponent dc;
        
        while (packet.hasMoreComponents()) {

            dc = packet.nextComponent();

            switch (dc.type) {
                /* 2002.06.05 MackyBee note: Temporary sample. */
                case HLProtocol.HTLS_DATA_NEWS_ARTICLE:
                    article = ((HLProtocol.NewsArticleComponent) dc).article;
                    break;
                default:
                    DebuggerOutput.debug("dispatchNewsArticle: " + 
                        "Unsupported dc.type: " + dc.type);
                    break;
            }

        }

        /* provisionally */
        return (article);
    }

    /* Task dispatch */
    /**
     *  Description of the Method
     *
     * @param  packet                       Description of the Parameter
     * @exception  IOException              Description of the Exception
     * @exception  HLTaskNotFoundException  Description of the Exception
     */
    void dispatchTask(HLProtocol.Packet packet) throws IOException,
        HLTaskNotFoundException {

        HLTask task = (HLTask) hlc.getTask(packet.header.trans);
        Object resultValue = null;

        try {
            if (packet.header.isError == 1) {

                /* In case of an error. */
                String error = new String("Unspecified error.");

               /* 2002.05.31 MackyBee modified */
                HLProtocol.DataComponent dh;
                
                while (packet.hasMoreComponents()) {
                    /* 2002.05.31 MackyBee modified */
                    dh = packet.nextComponent();
                    //HLProtocol.DataComponent dh = packet.nextComponent();

		// 2006-01-24 Gardien modified : error = new String(dh.data); -> error = new String(dh.data, hlc.encoding);
                    switch (dh.type) {
                        case HLProtocol.HTLS_DATA_TASKERROR:
                            error = new String(dh.data, hlc.encoding);
                            break;
                        default: 
                            break;
                    }
                }

                resultValue = error;

            } else {
                switch (task.type) {
                    case HLProtocol.HTLC_HDR_LOGIN:
                        dispatchLogin(packet);
                        break;
                    case HLProtocol.HTLC_HDR_DIR_LIST:
                        resultValue = dispatchFileList(packet);
                        break;
                    case HLProtocol.HTLC_HDR_USER_GETINFO:
                        resultValue = dispatchUserInfo(packet);
                        break;
                    case HLProtocol.HTLC_HDR_USER_LIST:
                        resultValue = dispatchUserList(packet);
                        break;
                    case HLProtocol.HTLC_HDR_NEWS_GET:
                        resultValue = dispatchNews(packet);
                        break;
                    case HLProtocol.HTLC_HDR_FILE_GET:
                        resultValue = dispatchFileGet(packet);
                        break;
                    case HLProtocol.HTLC_HDR_FILE_PUT:
//                        resultValue = dispatchFilePut(packet);
                        break;
                    case HLProtocol.HTLC_HDR_PRIVCHAT_CREATE:
                        resultValue = dispatchPrivateChatCreate(packet);
                        break;
                    case HLProtocol.HTLC_HDR_PRIVCHAT_JOIN:
                        resultValue = dispatchPrivateChatJoin(packet);
                        break;
                    case HLProtocol.HTLC_HDR_ACCOUNT_READ:
                        resultValue = dispatchAccountInfo(packet);
                        break;
                    case HLProtocol.HTLC_HDR_FILE_GETINFO:
                        resultValue = dispatchFileInfo(packet);
                        break;
                    /* 2002.05.31 MackyBee added */
                    case HLProtocol.HTLC_HDR_GET_NEWS_BUNDLE:
                        resultValue = dispatchNewsBundleList(packet);
                        break;
                    /* 2002.06.04 MackyBee added */
                    case HLProtocol.HTLC_HDR_GET_NEWS_THREADS:
                        resultValue = dispatchNewsThreadList(packet);
                        break;
                    /* 2002.06.05 MackyBee added */
                    case HLProtocol.HTLC_HDR_GET_NEWS_ARTICLE:
                        resultValue = dispatchNewsArticle(packet);
                        break;
                    default:

                        DebuggerOutput.debug("dispatchTask[" + packet.header.trans
                             + "] task [" + packet.header.trans + "] undispatched: "
                             + task);
                        break;
                }

            }

        } finally {

            /* Notify threads waiting on completion of this task. */
            if (packet.header.isError == 1) {
                task.setError(new HLTaskErrorException((String) resultValue));
            } else {
                task.setData(resultValue);
            }

            /* Either we dispose this task or someone else does it
               (i.e. HLClient.waitFor()). */
            if (task.disposeWhenReceived) {
                DebuggerOutput.debug("dispatchTask: task[" 
                    + packet.header.trans + "]: disposing task ...");
                hlc.disposeTask(packet.header.trans);
            }

            if (packet.header.isError == 1) {
                for (Enumeration en = hlc.hle.elements();
                    en.hasMoreElements(); ) {
                    ((HLClientListener) en.nextElement()).
                        handleTaskError(packet.header.trans, 
                        (String) resultValue);
                }
            } else {
                for (Enumeration en = hlc.hle.elements();
                    en.hasMoreElements(); ) {
                    ((HLClientListener) en.nextElement()).
                        handleTaskComplete(packet.header.trans);
                }
            }

            DebuggerOutput.debug("dispatchTask: task[" + packet.header.trans 
                + "]: dispatcher done.");
        }
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    Object dispatchFileInfo(HLProtocol.Packet packet) throws IOException {

        byte[] icon = new byte[0];
        byte[] type = new byte[0];
        byte[] creator = new byte[0];
        byte[] name = new byte[0];
        byte[] comment = new byte[0];
        long size = 0;
        Calendar created = Calendar.getInstance();
        Calendar modified = Calendar.getInstance();

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {

                case HLProtocol.HTLS_DATA_FILE_ICON:
                    icon = dh.data;
                    break;
                case HLProtocol.HTLS_DATA_FILE_TYPE:
                    type = dh.data;
                    break;
                case HLProtocol.HTLS_DATA_FILE_CREATOR:
                    creator = dh.data;
                    break;
                case HLProtocol.HTLS_DATA_FILE_NAME:
                    name = dh.data;
                    break;
                case HLProtocol.HTLS_DATA_FILE_COMMENT:
                    comment = dh.data;
                    break;
                case HLProtocol.HTLS_DATA_FILE_DATE_CREATED:
                    created = ((HLProtocol.DateComponent) dh).date;
                    break;
                case HLProtocol.HTLS_DATA_FILE_DATE_MODIFIED:
                    modified = ((HLProtocol.DateComponent) dh).date;
                    break;
                case HLProtocol.HTLS_DATA_FILE_SIZE:
                    size = ToArrayConverters.byteArrayToLong(dh.data);
                    break;
            }

        }

        HLProtocol.FileInfo info = hlp.new FileInfo();
        // 2002.12.30 MackyBee modified. 
        // new String(byte[]) -> new String(byte[], String)
        info.icon = new String(icon, hlc.encoding);
        info.type = new String(type, hlc.encoding);
        info.creator = new String(creator, hlc.encoding);
        info.name = new String(name, hlc.encoding);
        info.comment = new String(comment, hlc.encoding);
        info.size = size;
        info.created = created;
        info.modified = modified;

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handleFileInfo(packet.header.trans, info);
        }

        return info;
    }

    /* Accounts */
    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    Object dispatchAccountInfo(HLProtocol.Packet packet) throws IOException {

        byte[] nick = new byte[0];
        byte[] password = new byte[0];
        byte[] login = new byte[0];
        byte[] access2 = new byte[0];
        long access = 0;

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            if (dh.data != null) {
                switch (dh.type) {
                    case HLProtocol.HTLS_DATA_NICK:
                        nick = dh.data;
                        break;
                    case HLProtocol.HTLS_DATA_LOGIN:
                        login = HLProtocol.invert(dh.data);
                        break;
                    case HLProtocol.HTLS_DATA_PASSWORD:
                        password = HLProtocol.invert(dh.data);
                        break;
                    case HLProtocol.HTLS_DATA_PRIVILEGES:
                        access = ToArrayConverters.byteArrayToLong(
                            ToArrayConverters.swapByteArray(dh.data));
                        break;
                }
            }
        }

        HLProtocol.AccountInfo account = hlp.new AccountInfo(
            // 2002.12.30 MackyBee modified. Added encoding arguments.
            new String(login, hlc.encoding), 
            new String(nick, hlc.encoding), 
            new String(password, hlc.encoding), 
            access, null);

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handleAccountInfo(packet.header.trans, account);
        }

        return account;
    }

    /* Users */
    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    // 2002.12.21 MackyBee took NOTE: 
    // Needs process checking. 
    Object dispatchUserList(HLProtocol.Packet packet) throws IOException {

        Vector userV = new Vector();

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {
            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {
                case HLProtocol.HTLS_DATA_USER_LIST:
                    userV.insertElementAt((HLProtocol.UserListComponent) dh, 0);
                    break;
                default: 
                    break;
            }
        }

        HLProtocol.UserListComponent[] users =
            new HLProtocol.UserListComponent[userV.size()];

        for (int i = 0; i < userV.size(); i++) {
            users[i] = (HLProtocol.UserListComponent) userV.elementAt(i);
        }

        userV.removeAllElements();

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handleUserList(packet.header.trans, users);
        }

        return users;
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    Object dispatchUserInfo(HLProtocol.Packet packet) throws IOException {

        // 2002.12.28 MackyBee modified. 
        String userinfo = null;
        //String userinfo = new String();

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

		// 2006-01-24 Gardien modified : userinfo = new String(dh.data); -> userinfo = new String(dh.data, hlc.encoding);
            switch (dh.type) {
                case HLProtocol.HTLS_DATA_USER_INFO:
                    userinfo = new String(dh.data, hlc.encoding);
                    break;
            }
        }

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handleUserInfo(packet.header.trans, userinfo);
        }

        return userinfo;
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchUserChange(HLProtocol.Packet packet) throws IOException {

        int icon = 0;

        int sock = 0;

        int color = 0;
        /* 2002.03.22 MackyBee modified */
        /* 2002.04.04 MackyBee re-modified (returned to original code) */
        //byte[] nick = null;
        byte[] nick = new byte[0];

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {
            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {
                case HLProtocol.HTLS_DATA_SOCKET:
                    sock = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_ICON:
                    // Note 2002.03.18 by MackyBee
                    // icon should may be a short value.
                    icon = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
                case HLProtocol.HTLS_DATA_NICK:
                    nick = dh.data;
                    break;
                case HLProtocol.HTLS_DATA_COLOUR:
                    color = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
            }
        }

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handleUserChange(sock, 
                    // 2002.12.28 MackyBee modified. 
                    new String(nick, hlc.encoding), 
                    //new String(nick), 
                    icon, color);
        }
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchUserLeave(HLProtocol.Packet packet) throws IOException {

        int sock = 0;

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {
            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {
                case HLProtocol.HTLS_DATA_SOCKET:
                    sock = ToArrayConverters.byteArrayToInt(dh.data);
                    break;
            }
        }

        for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
            ((HLClientListener) en.nextElement()).
                handleUserLeave(sock);
        }
    }

    /* News post, get */
    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchNewsPost(HLProtocol.Packet packet) throws IOException {

        String post = null;

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {
            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {
                case HLProtocol.HTLC_DATA_NEWS_POST:
                    // 2002.12.28 MackyBee modified. 
                    post = new String(dh.data, hlc.encoding);
                    //post = new String(dh.data);
                    break;
            }
        }

        if (post != null) {
            for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
                ((HLClientListener) en.nextElement()).
                    handleNewsPost(post);
            }
        }
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @return                  Description of the Return Value
     * @exception  IOException  Description of the Exception
     */
    Object dispatchNews(HLProtocol.Packet packet) throws IOException {

        String news = null;

        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {
            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {
                case HLProtocol.HTLS_DATA_NEWS:
                    // 2002.12.28 MackyBee modified. 
                    news = new String(dh.data, hlc.encoding);
                    //news = new String(dh.data);
                    break;
            }
        }

        if (news != null) {
            for (Enumeration en = hlc.hle.elements(); en.hasMoreElements(); ) {
                ((HLClientListener) en.nextElement()).
                    handleNews(packet.header.trans, news);
            }
        }

        return news;
    }

    /* Agreement, login */
    /*
		Note 2002.02.28 MackyBee
		This method waits for agreement and user agree or disagree action.
	 */
    /**
     *  Description of the Method
     *
     * @exception  InterruptedException  Description of the Exception
     */
    synchronized void waitForAgreement() throws InterruptedException {

        if (agreementAborted) {
            throw new InterruptedException("agreement aborted");
        }

        DebuggerOutput.debug("HLClientDispatcher: waiting for agreement ...");
        //while(!agreementReceived && !agreementAborted)
        /* Modified 2002.03.02 MackyBee */
        //waitingAgree = true;
        if (!agreementReceived && !agreementAborted) {
            waitingAgree = true;
            //wait();
            /* 	Experimentally 2002.03.04 MackyBee
					On version 1.2.3 server, client can't detect if there is
					agreement by stream protocol.
					Then, simply wait a while for agreement coming.
					If agreement doesn't come, starts login process.
				 */

            // 2002.09.29 MackyBee deactivated the wait method call.
            // Now, process for 1.2.3 version has not been used here.
            /* Modified the time length 1000L -> 3000L
				   on 2002.03.21 by MackyBee
                   -> 2000L on 2002.03.25 by MackyBee */
            //wait(2000L);
            if (waitNoTimeLimit) {
                wait();
            }
            waitingAgree = false;
        }
        //waitingAgree = false;
        //}
        DebuggerOutput.debug("HLClientDispatcher.agreementReceived: "
             + agreementReceived);
        DebuggerOutput.debug("HLClientDispatcher.agreementAborted: "
             + agreementAborted);
        DebuggerOutput.debug("HLClientDispatcher.agreementAccepted: "
             + agreementAccepted);

        DebuggerOutput.debug("HLClientDispatcher: got agreement.");

        if (agreementAborted) {
            throw new InterruptedException("agreement aborted");
        }
    }

    /**  Description of the Method */
    synchronized void notifyAgreementAborted() {
        agreementAborted = true;
        notify();
    }

    /**  Description of the Method */
    synchronized void notifyAgreementReceived() {
        // TEMPORARY
        //System.out.println("Entered to notifyAgreementReceived()...");

        if (!agreementAborted) {

            agreementReceived = true;
            waitNoTimeLimit = false;

            /* Experimentally added 2002.03.01 by MackyBee */
            if (waitingAgree) {
                DebuggerOutput.debug(
                    "HLClientDispatcher: Notify agreement received.");

                notify();
            }
        }

        // 2002.09.29 MackyBee added.
        if (!agreementAccepted) {
            //try {
                for (Enumeration en = hlc.hle.elements(); 
                    en.hasMoreElements(); ) {
                    ((HLClientListener) en.nextElement()).
                    handleDisconnect(null);
                }
                //hlc.setConnected(false);
                //hlc.close();
                // PROVISIONALLY
                //System.out.println("CLOSE! [agreementAccepted == false]");
            //} catch (IOException e) {
                // PROVISIONALLY
            //    System.out.println("Exception during hlc.close()");
            //}
        }

        // 2002.09.29 TEMPORARY
        //System.out.println("second agreementAccepted: " + agreementAccepted);
        //System.out.println("Existing from notifyAgreementReceived()...");
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchAgreement(HLProtocol.Packet packet) throws IOException {

        try {
            /* Added 2002.03.03 MackyBee */
            if (packet.header.hc == 0) {
                /* Treats as accepted */
                DebuggerOutput.debug("HLClientDispatcher.dispatchAgreement: " 
                    + "No agreement, treats as accepted.");
                agreementAccepted = true;
            }

            /* 2002.05.31 MackyBee modified */
            HLProtocol.DataComponent dh;
            
            while (packet.hasMoreComponents()) {
    
                /* 2002.05.31 MackyBee modified */
                dh = packet.nextComponent();
                //HLProtocol.DataComponent dh = packet.nextComponent();

                switch (dh.type) {

                    case HLProtocol.HTLS_DATA_AGREEMENT:

                        /* Added 2002.03.04 MackyBee */
                        waitNoTimeLimit = true;

                        // 2002.12.28 MackyBee modified. 
                        String agreement = new String(dh.data, hlc.encoding);
                        //String agreement = new String(dh.data);

                        for (Enumeration en = hlc.hle.elements(); 
                            en.hasMoreElements(); ) {

                            /* Added for testing 2002.02.28 MackyBee */
                            DebuggerOutput.debug("HLClientDispatcher: " +
                                "Agreement being handled...");

                            // 2002.09.29 modified to refer Machine class.
                            // This change may cause package dependency.
                            // Modified 2002.02.28 MackyBee
                            HLClientListener hlcl = 
                                (HLClientListener) en.nextElement();
//                            if (hlcl instanceof pharerouge.client.Machine) {
//                                agreementAccepted = 
//                                    hlcl.handleAgreement(agreement);
//                            }
                            //    ((HLClientListener) en.nextElement()).
                            //    handleAgreement(agreement);

                            // 2002.09.29 TEMPORARY
                            //System.out.println("agreementAccepted: " 
                            //    + agreementAccepted);
                        }

                        break;
                    case HLProtocol.HTLS_DATA_AUTO_AGREE:

                        /* Added for testing 2002.02.28 MackyBee */
                        DebuggerOutput.debug("HLClientDispatcher: " +
                            "Agreement was automatically agreed.");

                        agreementAccepted = true;
                        int autoAgree = ToArrayConverters.byteArrayToInt(
                            dh.data);
                        break;
                    // 2002.09.12 MackyBee added. Almost no meaning.
                    default:
                        break;
                }

            }

        } finally {
            // 2002.09.29 TEMPORARY added
            //System.out.println("notifyAgreementReceived()");

            notifyAgreementReceived();
        }
    }

    /**
     *  Description of the Method
     *
     * @param  packet           Description of the Parameter
     * @exception  IOException  Description of the Exception
     */
    void dispatchLogin(HLProtocol.Packet packet) throws IOException {

        /* Added experimentally 2002.02.15 by MackyBee */
        DebuggerOutput.debug("HLClientDispather.dispatchLogin...");

        //{{{ old process
        //int version = 0;

        /* Added 2002.03.02 MackyBee */
        /* 	If packet.header.hc == 0, it means version 1.2.3 server
			doesn't have agreement.
			Note 2002.03.03
			This approach is not suitable. I'll think another way.
			*/
        /*
		if ((packet.header.hc == 0) && (!agreementReceived)) {
			DebuggerOutput.debug("HLClientDispather.dispatchLogin " +
				"Detected as the server doesn't have agreement.");
			agreementAccepted = true;
			notifyAgreementReceived();
			// packet.hasMoreComponents() is none, then return here
			return;
		}
		*/
        //}}}
        
        /* 2002.05.31 MackyBee modified */
        HLProtocol.DataComponent dh;
        
        while (packet.hasMoreComponents()) {

            /* 2002.05.31 MackyBee modified */
            dh = packet.nextComponent();
            //HLProtocol.DataComponent dh = packet.nextComponent();

            switch (dh.type) {

                case HLProtocol.HTLS_DATA_VERSION:
                    /* Modified 2002.03.02 MackyBee */
                    //version = ToArrayConverters.byteArrayToInt(dh.data);
                    hlc.setServerVersion(ToArrayConverters.byteArrayToInt(
                        dh.data));
                    break;
                /* Added 2002.02.28 MackyBee */
                case HLProtocol.HTLS_DATA_SERVER_NAME:
                    DebuggerOutput.debug("HLClientDispather.dispatchLogin " +
                        "Server name was detected.");
                    for (Enumeration en = hlc.hle.elements();
                        en.hasMoreElements(); ) {
                        ((HLClientListener) en.nextElement()).handleServerName(
                            // 2002.12.28 MackyBee modified. 
                            new String(dh.data, hlc.encoding));
                            //new String(dh.data));
                    }
                    break;
                default:
                    DebuggerOutput.debug("dispatchLogin: unknown dh.type: " 
                        + dh.type);
                    break;
            }
        }

        //hlc.setServerVersion(version);
    }
}

