/*
	PeriodicalPacketSender.java
	
	Copyright(C) 2002 MackyBee <macky@email.plala.or.jp>
	
	This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package pharerouge.hotline;

import java.io.IOException;

/**
	This class sends a packet every specified interval in a Thread.
    It enables same function as offcial HLC 1.8.5's periodical (3 minutes) 
    "User list request" while idling time.
 */
public class PeriodicalPacketSender extends Thread {
	
	boolean destroyOrder = false; /* 2002.03.22 MackyBee added */
	long waitingTime;
	HLClient hlc;
	int type;
	HLProtocol.DataComponent[] dataComponents;
	
	public PeriodicalPacketSender(
		HLClient hlc, 
		int type, 
		HLProtocol.DataComponent[] dataComponents, 
		long time) {

		this.hlc = hlc;
		this.type = type;
		this.dataComponents = dataComponents;
		
		waitingTime = time;
	}
	
	void setWait(long time) {
		if (!isAlive()) { 
			return;
		}
		waitingTime = time;
		this.interrupt();
	}
	
	/* 2002.03.22 MackyBee added */
	public void destroy() {
		destroyOrder = true;
		this.interrupt();
	}
	
	/* 2002.03.22 MackyBee modified the while's condition */
	public void run() {
		while (hlc.connected && !destroyOrder) {
			try {
				sleep(waitingTime);
				/* Write here the periodical process */
				executeProcess();
			} catch (InterruptedException e) {
				/* In fact, this is not a Exception */
				
				/* Provisional support */
				//System.out.println("PeriodicalPacketSender: " + 
				//	"reset waitingTime.");
			}
		}
	}
	
	private void executeProcess() {
        // 2002.10.14 MackyBee modified if condition.
		// 2002.03.22 MackyBee modified the if's condition 
        if (!hlc.connected || destroyOrder) {
		//if (!hlc.connected && destroyOrder) {
			return;
		}
		HLProtocol.Packet packet = hlc.hlp.new Packet(
			type,
			hlc.nextTrans(),
			dataComponents);
		try {
			HLTask task = hlc.createTask(packet, true);
			packet.write(hlc.output);
		} catch (IOException e) {
            // 2002.10.14 MackyBee modified.
            e.printStackTrace();
			// Provisional support
			//System.out.println("PeriodicalPacketSender: " + e.getMessage());
		}
	}
}
