/**
 *  $Id: DebuggerOutput.java,v 1.1.1.1 2001/07/22 02:44:49 groomed Exp $
 *  Copyright (C) 1998-2001 groomed <groomed@users.sourceforge.net> This program
 *  is free software; you can redistribute it and/or modify it under the terms
 *  of the GNU General Public License as published by the Free Software
 *  Foundation; either version 2 of the License, or (at your option) any later
 *  version. This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 *  more details. You should have received a copy of the GNU General Public
 *  License along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

package pharerouge.utils;

// 2002.10.25 MackyBee replaced.
/* 2002.04.26 MackyBee added */
import java.util.HashMap;
import java.util.Map;

/* 2002.04.18 MackyBee modified */
/**  Some stuff for debugging.  */
public final class DebuggerOutput {
//public class DebuggerOutput {

    /**  Description of the Field */
    public static boolean on = true;

    // 2002.10.25 MackyBee replaced.
    /* 2002.04.26 MackyBee experimentally added */
    private static Map timeStampTable;
    //private static Hashtable timeStampTable = new Hashtable();

    /* 2002.06.19 MackyBee experimentally added. */
    private static boolean stackTraceInvoked = false;
    
    /**
     *  Conditionally writes the specified string to standard output.
     *
     * @param  s  a message.
     */
    public final static void debug(String s) {
        if (on) {
            System.out.println(s);
        }
    }

    /**
     * Conditionally dumps stack trace for the specified exception to
     * 'System.out'.
     *
     * @param  e  the exception.
     */
    public final static void stackTrace(Exception e) {
        // 2002.08.26 MackyBee changed specification.

        if (on) {
            e.printStackTrace(System.out);
        }

        /* 2002.06.19 MackyBee added. */
        //stackTraceInvoked = true;
    }

    /* 2002.08.26 deactivated. 2002.06.19 MackyBee added. */
    /*
    public static boolean isStackTraceInvoked() {
        return (stackTraceInvoked);
    }
    */

    // 2002.10.25 MackyBee deactivated.
    /* 2002.04.26 MackyBee added */
    /*
    public static Object getTimeStamp(Object key) {
        return (timeStampTable.get(key));
    }
    public static void setTimeStamp(Object key, Object value) {
        timeStampTable.put(key, value);
    }
    */

    /** 2002.07.12 MackyBee added. */
    public static void setTimeCount(Object key) {
        // 2002.10.25 MackyBee modified.
        if (timeStampTable == null) {
            timeStampTable = new HashMap();
        }
        timeStampTable.put(key, new Long(System.currentTimeMillis()));
    }
    /** 2002.07.12 MackyBee added. */
    public static long getTimeCount(Object key) {
        // 2002.10.25 MackyBee modified.
        if (timeStampTable != null) {
            return (System.currentTimeMillis() - 
                ((Long) timeStampTable.get(key)).longValue());
        } else {
            return (-1L);
        }
    }

    /* 2002.05.02 MackyBee added */
    public final static void putHexString(byte[] b) {
        if (on) {
            StringBuffer sb = new StringBuffer();
            sb.append("DebuggerOutput: to Hex Strings [");
            for (int i = 0; i < b.length; i++) {
                sb.append(Integer.toHexString((int) b[i]));
                sb.append(" ");
            }
            sb.setLength(sb.length() - 1);
            sb.append("]");
            System.out.println(sb.toString());
        }
    }
    
    /**
     *  Conditionally dumps a stack trace to standard output.
     *
     * @param  s  a message.
     */
    public final static void tell(String s) {
        if (on) {
            try {
                throw new DebuggerException(s);
            } catch (DebuggerException e) {
                e.printStackTrace();
            }
        }
    }

    /**
     *  Conditionally produces a hexdump of the specified byte array on standard
     *  output. The cols variable controls the number of horizontal columns.
     *
     * @param  b     the byte array to dump.
     * @param  cols  the number of horizontal columns.
     */
    public final static void dumpBytes(byte[] b, int cols) {

        if (on) {

            for (int i = 0; i < (b.length / cols) + 1; i++) {

                int offset = cols;

                for (int j = 0; j < cols && j + (i * cols) < b.length; offset--, j++) {
                    System.out.print(TextUtils.prepad(Integer.toHexString(b[j + (i * cols)] & 0xff), 1) + " ");
                }

                if (offset > 0) {
                    for (; offset > 0; offset--) {
                        System.out.print("   ");
                    }
                }

                System.out.print(" ");

                for (int j = 0; j < cols && j + (i * cols) < b.length; j++) {

                    if (!Character.isISOControl((char) b[j + (i * cols)])) {
                        System.out.print((char) b[j + (i * cols)]);
                    } else {

                        System.out.print(".");
                    }

                }
                System.out.println();
            }
        }
    }

    /**
     *  Enables or disables output from the methods in this class.
     *
     * @param  enable  true to enable output.
     */
    public static void setEnabled(boolean enable) {
        on = enable;

        if (on) {
            debug("Debugging output turned on");
        }
    }

}

/**  Thrown by DebuggerOutput.tell.  */
class DebuggerException extends Exception {

    /**
     *  Constructor for the DebuggerException object
     *
     * @param  s  Description of the Parameter
     */
    DebuggerException(String s) {
        super(s);
    }

}

